import logging
import socket
import xml.etree.ElementTree as ET
from ipaddress import IPv4Network

import ifaddr

from .security import LocalSecurity

_LOGGER = logging.getLogger(__name__)

BROADCAST_MSG = bytearray(
    [
        0x5A,
        0x5A,
        0x01,
        0x11,
        0x48,
        0x00,
        0x92,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x7F,
        0x75,
        0xBD,
        0x6B,
        0x3E,
        0x4F,
        0x8B,
        0x76,
        0x2E,
        0x84,
        0x9C,
        0x6E,
        0x57,
        0x8D,
        0x65,
        0x90,
        0x03,
        0x6E,
        0x9D,
        0x43,
        0x42,
        0xA5,
        0x0F,
        0x1F,
        0x56,
        0x9E,
        0xB8,
        0xEC,
        0x91,
        0x8E,
        0x92,
        0xE5,
    ]
)

DEVICE_INFO_MSG = bytearray(
    [
        0x5A,
        0x5A,
        0x15,
        0x00,
        0x00,
        0x38,
        0x00,
        0x04,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x27,
        0x33,
        0x05,
        0x13,
        0x06,
        0x14,
        0x14,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x03,
        0xE8,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0xCA,
        0x8D,
        0x9B,
        0xF9,
        0xA0,
        0x30,
        0x1A,
        0xE3,
        0xB7,
        0xE4,
        0x2D,
        0x53,
        0x49,
        0x47,
        0x62,
        0xBE,
    ]
)


def discover(discover_type=None, ip_address=None):
    if discover_type is None:
        discover_type = []
    security = LocalSecurity()
    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
    sock.settimeout(5)
    found_devices = {}
    if ip_address is None:
        addrs = enum_all_broadcast()
    else:
        addrs = [ip_address]
    _LOGGER.debug(f"All addresses for broadcast: {addrs}")
    for addr in addrs:
        try:
            sock.sendto(BROADCAST_MSG, (addr, 6445))
            sock.sendto(BROADCAST_MSG, (addr, 20086))
        except Exception:
            _LOGGER.warning(f"Can't access network {addr}")
    while True:
        try:
            data, addr = sock.recvfrom(512)
            ip = addr[0]
            _LOGGER.debug(f"Received response from {addr}: {data.hex()}")
            if len(data) >= 104 and (
                data[:2].hex() == "5a5a" or data[8:10].hex() == "5a5a"
            ):
                if data[:2].hex() == "5a5a":
                    protocol = 2
                elif data[:2].hex() == "8370":
                    protocol = 3
                    if data[8:10].hex() == "5a5a":
                        data = data[8:-16]
                else:
                    continue
                device_id = int.from_bytes(
                    bytearray.fromhex(data[20:26].hex()), "little"
                )
                if device_id in found_devices:
                    continue
                encrypt_data = data[40:-16]
                reply = security.aes_decrypt(encrypt_data)
                _LOGGER.debug(f"Declassified reply: {reply.hex()}")
                ssid = reply[41 : 41 + reply[40]].decode("utf-8")
                device_type = ssid.split("_")[1]
                port = bytes2port(reply[4:8])
                model = reply[17:25].decode("utf-8")
                sn = reply[8:40].decode("utf-8")
            elif data[:6].hex() == "3c3f786d6c20":
                protocol = 1
                root = ET.fromstring(data.decode(encoding="utf-8", errors="replace"))
                child = root.find("body/device")
                m = child.attrib
                port, sn, device_type = (
                    int(m["port"]),
                    m["apc_sn"],
                    str(hex(int(m["apc_type"])))[2:],
                )
                response = get_device_info(ip, int(port))
                device_id = get_id_from_response(response)
                if len(sn) == 32:
                    model = sn[9:17]
                elif len(sn) == 22:
                    model = sn[3:11]
                else:
                    model = ""
            else:
                continue
            device = {
                "device_id": device_id,
                "type": int(device_type, 16),
                "ip_address": ip,
                "port": port,
                "model": model,
                "sn": sn,
                "protocol": protocol,
            }
            if len(discover_type) == 0 or device.get("type") in discover_type:
                found_devices[device_id] = device
                _LOGGER.debug(f"Found a supported device: {device}")
            else:
                _LOGGER.debug(f"Found a unsupported device: {device}")
        except socket.timeout:
            break
        except socket.error as e:
            _LOGGER.error(f"Socket error: {repr(e)}")
    return found_devices


def get_id_from_response(response):
    if response[64:-16][:6].hex() == "3c3f786d6c20":
        xml = response[64:-16]
        root = ET.fromstring(xml.decode(encoding="utf-8", errors="replace"))
        child = root.find("smartDevice")
        m = child.attrib
        return int.from_bytes(bytearray.fromhex(m["devId"]), "little")
    else:
        return 0


def bytes2port(paramArrayOfbyte):
    if paramArrayOfbyte is None:
        return 0
    b, i = 0, 0
    while b < 4:
        if b < len(paramArrayOfbyte):
            b1 = paramArrayOfbyte[b] & 0xFF
        else:
            b1 = 0
        i |= b1 << b * 8
        b += 1
    return i


def get_device_info(device_ip, device_port: int):
    response = bytearray(0)
    try:
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
            sock.settimeout(8)
            device_address = (device_ip, device_port)
            sock.connect(device_address)
            _LOGGER.debug(
                f"Sending to {device_ip}:{device_port} {DEVICE_INFO_MSG.hex()}"
            )
            sock.sendall(DEVICE_INFO_MSG)
            response = sock.recv(512)
    except socket.timeout:
        _LOGGER.warning(
            f"Connect the device {device_ip}:{device_port} timed out for 8s. "
            f"Don't care about a small amount of this. if many maybe not support."
        )
    except socket.error:
        _LOGGER.warning(f"Can't connect to Device {device_ip}:{device_port}")
    return response


def enum_all_broadcast():
    nets = []
    adapters = ifaddr.get_adapters()
    for adapter in adapters:
        for ip in adapter.ips:
            if ip.is_IPv4 and ip.network_prefix < 32:
                local_network = IPv4Network(
                    f"{ip.ip}/{ip.network_prefix}", strict=False
                )
                if (
                    local_network.is_private
                    and not local_network.is_loopback
                    and not local_network.is_link_local
                ):
                    addr = str(local_network.broadcast_address)
                    if addr not in nets:
                        nets.append(addr)
    return nets



from ...message import MessageType, MessageRequest, MessageResponse, MessageBody


class MessageB0Base(MessageRequest):
    def __init__(self, protocol_version, message_type, body_type):
        super().__init__(
            device_type=0xB0,
            protocol_version=protocol_version,
            message_type=message_type,
            body_type=body_type,
        )

    @property
    def _body(self):
        raise NotImplementedError


class MessageQuery00(MessageB0Base):
    def __init__(self, protocol_version):
        super().__init__(
            protocol_version=protocol_version,
            message_type=MessageType.query,
            body_type=0x00,
        )

    @property
    def _body(self):
        return bytearray([])


class MessageQuery01(MessageB0Base):
    def __init__(self, protocol_version):
        super().__init__(
            protocol_version=protocol_version,
            message_type=MessageType.query,
            body_type=0x01,
        )

    @property
    def _body(self):
        return bytearray([])


class B0MessageBody(MessageBody):
    def __init__(self, body):
        super().__init__(body)
        if len(body) > 15:
            self.door = (body[0] & 0x80) > 0
            self.status = body[0] & 0x7F
            self.time_remaining = body[2] * 60 + body[3]
            self.error_code = body[5]


class B0Message01Body(MessageBody):
    def __init__(self, body):
        super().__init__(body)
        if len(body) > 15:
            self.door = (body[32] & 0x02) > 0
            self.status = body[31]
            self.time_remaining = (
                (0 if body[22] == 0xFF else body[22]) * 3600
                + (0 if body[23] == 0xFF else body[23]) * 60
                + (0 if body[24] == 0xFF else body[24])
            )
            self.current_temperature = (body[25] << 8) + (body[26])
            if self.current_temperature == 0:
                self.current_temperature = (body[27] << 8) + body[28]
            self.tank_ejected = (body[32] & 0x04) > 0
            self.water_shortage = (body[32] & 0x08) > 0
            self.water_change_reminder = (body[32] & 0x10) > 0


class MessageB0Response(MessageResponse):
    def __init__(self, message):
        super().__init__(message)
        if self.message_type in [MessageType.notify1, MessageType.query]:
            if self.body_type == 0x01:
                self.set_body(B0Message01Body(super().body))
            elif self.body_type == 0x04:
                pass
            else:
                self.set_body(B0MessageBody(super().body))
        self.set_attr()



import logging
import json
from .message import MessageQuery, MessageB6Response, MessageSet

try:
    from enum import StrEnum
except ImportError:
    from ...backports.enum import StrEnum
from ...device import MideaDevice

_LOGGER = logging.getLogger(__name__)


class DeviceAttributes(StrEnum):
    power = "power"
    light = "light"
    mode = "mode"
    fan_level = "fan_level"
    fan_speed = "fan_speed"
    oilcup_full = "oilcup_full"
    cleaning_reminder = "cleaning_reminder"


class MideaB6Device(MideaDevice):
    def __init__(
        self,
        name: str,
        device_id: int,
        ip_address: str,
        port: int,
        token: str,
        key: str,
        protocol: int,
        model: str,
        subtype: int,
        customize: str,
    ):
        super().__init__(
            name=name,
            device_id=device_id,
            device_type=0xB6,
            ip_address=ip_address,
            port=port,
            token=token,
            key=key,
            protocol=protocol,
            model=model,
            subtype=subtype,
            attributes={
                DeviceAttributes.power: False,
                DeviceAttributes.light: None,
                DeviceAttributes.mode: None,
                DeviceAttributes.fan_level: 0,
                DeviceAttributes.fan_speed: 0,
                DeviceAttributes.oilcup_full: False,
                DeviceAttributes.cleaning_reminder: False,
            },
        )
        self._default_speeds = {0: "Off", 1: "Level 1", 2: "Level 2"}
        self._default_power_speed = 2
        self._power_speed = self._default_power_speed
        self._speeds = self._default_speeds
        self.set_customize(customize)

    @property
    def speed_count(self):
        return len(self._speeds) - 1

    @property
    def preset_modes(self):
        return list(self._speeds.values())

    def build_query(self):
        return [MessageQuery(self._protocol_version)]

    def process_message(self, msg):
        message = MessageB6Response(msg)
        self._protocol_version = message.protocol_version
        _LOGGER.debug(f"[{self.device_id}] Received: {message}")
        new_status = {}
        for status in self._attributes.keys():
            if hasattr(message, str(status)):
                value = getattr(message, str(status))
                if status == DeviceAttributes.fan_level:
                    if value in self._speeds.keys():
                        self._attributes[DeviceAttributes.mode] = self._speeds.get(
                            value
                        )
                        self._attributes[DeviceAttributes.fan_speed] = list(
                            self._speeds.keys()
                        ).index(value)
                    else:
                        self._attributes[DeviceAttributes.mode] = None
                        self._attributes[DeviceAttributes.fan_speed] = 0
                    new_status[DeviceAttributes.mode.value] = self._attributes[
                        DeviceAttributes.mode
                    ]
                    new_status[DeviceAttributes.fan_speed.value] = self._attributes[
                        DeviceAttributes.fan_speed
                    ]
                self._attributes[status] = getattr(message, str(status))
                new_status[str(status)] = self._attributes[status]
        return new_status

    def set_attribute(self, attr, value):
        message = None
        if attr == DeviceAttributes.fan_speed:
            if value < len(self._speeds):
                message = MessageSet(self._protocol_version)
                message.fan_level = list(self._speeds.keys())[value]
        elif attr == DeviceAttributes.mode:
            if value in self._speeds.values():
                message = MessageSet(self._protocol_version)
                message.fan_level = list(self._speeds.keys())[
                    list(self._speeds.values()).index(value)
                ]
            elif not value:
                message = MessageSet(self._protocol_version)
                message.power = False
        elif attr == DeviceAttributes.power:
            message = MessageSet(self._protocol_version)
            message.power = value
            message.fan_level = self._power_speed
        elif attr == DeviceAttributes.light:
            message = MessageSet(self._protocol_version)
            message.light = value
        if message is not None:
            self.build_send(message)

    def turn_on(self, fan_speed=None, mode=None):
        message = MessageSet(self._protocol_version)
        message.power = True
        if fan_speed is not None and fan_speed < len(self._speeds):
            message.fan_level = list(self._speeds.keys())[fan_speed]
        else:
            message.fan_level = self._power_speed
        if mode is not None in self._speeds.values():
            message.fan_level = list(self._speeds.keys())[
                list(self._speeds.values()).index(mode)
            ]
        self.build_send(message)

    def set_customize(self, customize):
        self._speeds = self._default_speeds
        self._power_speed = self._default_power_speed
        if customize and len(customize) > 0:
            try:
                params = json.loads(customize)
                if params:
                    if "default_speed" in params:
                        self._power_speed = int(params.get("default_speed"))
                    if "speeds" in params:
                        self._speeds = {}
                        speeds = {}
                        for k, v in params.get("speeds").items():
                            speeds[int(k)] = v
                        keys = sorted(speeds.keys())
                        for k in keys:
                            self._speeds[k] = speeds[k]
                    self.update_all(
                        {"speeds": self._speeds, "default_speed": self._power_speed}
                    )
            except Exception as e:
                _LOGGER.error(f"[{self.device_id}] Set customize error: {repr(e)}")


class MideaAppliance(MideaB6Device):
    pass



import logging
import socket
import threading
import time
from enum import IntEnum

from .backports.enum import StrEnum
from .message import (
    MessageApplianceResponse,
    MessageQueryAppliance,
    MessageQuestCustom,
    MessageType,
)
from .packet_builder import PacketBuilder
from .security import (
    MSGTYPE_ENCRYPTED_REQUEST,
    MSGTYPE_HANDSHAKE_REQUEST,
    LocalSecurity,
)

_LOGGER = logging.getLogger(__name__)


class AuthException(Exception):
    pass


class ResponseException(Exception):
    pass


class RefreshFailed(Exception):
    pass


class DeviceAttributes(StrEnum):
    pass


class ParseMessageResult(IntEnum):
    SUCCESS = 0
    PADDING = 1
    ERROR = 99


class MideaDevice(threading.Thread):
    def __init__(
        self,
        name: str,
        device_id: int,
        device_type: int,
        ip_address: str,
        port: int,
        token: str,
        key: str,
        protocol: int,
        model: str,
        subtype: int,
        attributes: dict,
    ):
        threading.Thread.__init__(self)
        self._attributes = attributes if attributes else {}
        self._socket = None
        self._ip_address = ip_address
        self._port = port
        self._security = LocalSecurity()
        self._token = bytes.fromhex(token) if token else None
        self._key = bytes.fromhex(key) if key else None
        self._buffer = b""
        self._device_name = name
        self._device_id = device_id
        self._device_type = device_type
        self._protocol = protocol
        self._model = model
        self._subtype = subtype
        self._protocol_version = 0
        self._updates = []
        self._unsupported_protocol = []
        self._is_run = False
        self._available = True
        self._appliance_query = True
        self._refresh_interval = 30
        self._heartbeat_interval = 10
        self._default_refresh_interval = 30

    @property
    def name(self):
        return self._device_name

    @property
    def available(self):
        return self._available

    @property
    def device_id(self):
        return self._device_id

    @property
    def device_type(self):
        return self._device_type

    @property
    def model(self):
        return self._model

    @property
    def subtype(self):
        return self._subtype

    @staticmethod
    def fetch_v2_message(msg):
        result = []
        while len(msg) > 0:
            factual_msg_len = len(msg)
            if factual_msg_len < 6:
                break
            alleged_msg_len = msg[4] + (msg[5] << 8)
            if factual_msg_len >= alleged_msg_len:
                result.append(msg[:alleged_msg_len])
                msg = msg[alleged_msg_len:]
            else:
                break
        return result, msg

    def connect(self, refresh_status=True):
        try:
            self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self._socket.settimeout(10)
            _LOGGER.debug(
                f"[{self._device_id}] Connecting to {self._ip_address}:{self._port}"
            )
            self._socket.connect((self._ip_address, self._port))
            _LOGGER.debug(f"[{self._device_id}] Connected")
            if self._protocol == 3:
                self.authenticate()
            _LOGGER.debug(f"[{self._device_id}] Authentication success")
            if refresh_status:
                self.refresh_status(wait_response=True)
            self.enable_device(True)
            return True
        except socket.timeout:
            _LOGGER.debug(f"[{self._device_id}] Connection timed out")
        except socket.error:
            _LOGGER.debug(f"[{self._device_id}] Connection error")
        except AuthException:
            _LOGGER.debug(f"[{self._device_id}] Authentication failed")
        except ResponseException:
            _LOGGER.debug(f"[{self._device_id}] Unexpected response received")
        except RefreshFailed:
            _LOGGER.debug(f"[{self._device_id}] Refresh status is timed out")
        except Exception as e:
            _LOGGER.error(
                f"[{self._device_id}] Unknown error: {e.__traceback__.tb_frame.f_globals['__file__']}, "
                f"{e.__traceback__.tb_lineno}, {repr(e)}"
            )
        self.enable_device(False)
        return False

    def authenticate(self):
        request = self._security.encode_8370(self._token, MSGTYPE_HANDSHAKE_REQUEST)
        _LOGGER.debug(f"[{self._device_id}] Handshaking")
        self._socket.send(request)
        response = self._socket.recv(512)
        if len(response) < 20:
            raise AuthException()
        response = response[8:72]
        self._security.tcp_key(response, self._key)

    def send_message(self, data):
        if self._protocol == 3:
            self.send_message_v3(data, msg_type=MSGTYPE_ENCRYPTED_REQUEST)
        else:
            self.send_message_v2(data)

    def send_message_v2(self, data):
        if self._socket is not None:
            self._socket.send(data)
        else:
            _LOGGER.debug(
                f"[{self._device_id}] Send failure, device disconnected, data: {data.hex()}"
            )

    def send_message_v3(self, data, msg_type=MSGTYPE_ENCRYPTED_REQUEST):
        data = self._security.encode_8370(data, msg_type)
        self.send_message_v2(data)

    def build_send(self, cmd):
        data = cmd.serialize()
        _LOGGER.debug(f"[{self._device_id}] Sending: {cmd}")
        msg = PacketBuilder(self._device_id, data).finalize()
        self.send_message(msg)

    def refresh_status(self, wait_response=False):
        cmds: list = self.build_query()
        if self._appliance_query:
            cmds = [MessageQueryAppliance(self.device_type)] + cmds
        error_count = 0
        for cmd in cmds:
            if cmd.__class__.__name__ not in self._unsupported_protocol:
                self.build_send(cmd)
                if wait_response:
                    try:
                        while True:
                            msg = self._socket.recv(512)
                            if len(msg) == 0:
                                raise socket.error
                            result = self.parse_message(msg)
                            if result == ParseMessageResult.SUCCESS:
                                break
                            elif result == ParseMessageResult.PADDING:
                                continue
                            else:
                                raise ResponseException
                    except socket.timeout:
                        error_count += 1
                        self._unsupported_protocol.append(cmd.__class__.__name__)
                        _LOGGER.debug(
                            f"[{self._device_id}] Does not supports "
                            f"the protocol {cmd.__class__.__name__}, ignored"
                        )
                    except ResponseException:
                        error_count += 1
            else:
                error_count += 1
        if error_count == len(cmds):
            raise RefreshFailed

    def pre_process_message(self, msg):
        if msg[9] == MessageType.query_appliance:
            message = MessageApplianceResponse(msg)
            self._appliance_query = False
            _LOGGER.debug(f"[{self.device_id}] Received: {message}")
            self._protocol_version = message.protocol_version
            _LOGGER.debug(
                f"[{self._device_id}] Device protocol version: {self._protocol_version}"
            )
            return False
        return True

    def parse_message(self, msg):
        if self._protocol == 3:
            messages, self._buffer = self._security.decode_8370(self._buffer + msg)
        else:
            messages, self._buffer = self.fetch_v2_message(self._buffer + msg)
        if len(messages) == 0:
            return ParseMessageResult.PADDING
        for message in messages:
            if message == b"ERROR":
                return ParseMessageResult.ERROR
            payload_len = message[4] + (message[5] << 8) - 56
            payload_type = message[2] + (message[3] << 8)
            if payload_type in [0x1001, 0x0001]:
                # Heartbeat detected
                pass
            elif len(message) > 56:
                cryptographic = message[40:-16]
                if payload_len % 16 == 0:
                    decrypted = self._security.aes_decrypt(cryptographic)
                    try:
                        cont = True
                        if self._appliance_query:
                            cont = self.pre_process_message(decrypted)
                        if cont:
                            status = self.process_message(decrypted)
                            if len(status) > 0:
                                self.update_all(status)
                            else:
                                _LOGGER.debug(
                                    f"[{self._device_id}] Unidentified protocol"
                                )
                    except Exception:
                        _LOGGER.error(
                            f"[{self._device_id}] Error in process message, msg = {decrypted.hex()}"
                        )
                else:
                    _LOGGER.warning(
                        f"[{self._device_id}] Illegal payload, "
                        f"original message = {msg.hex()}, buffer = {self._buffer.hex()}, "
                        f"8370 decoded = {message.hex()}, payload type = {payload_type}, "
                        f"alleged payload length = {payload_len}, factual payload length = {len(cryptographic)}"
                    )
            else:
                _LOGGER.warning(
                    f"[{self._device_id}] Illegal message, "
                    f"original message = {msg.hex()}, buffer = {self._buffer.hex()}, "
                    f"8370 decoded = {message.hex()}, payload type = {payload_type}, "
                    f"alleged payload length = {payload_len}, message length = {len(message)}, "
                )
        return ParseMessageResult.SUCCESS

    def build_query(self):
        raise NotImplementedError

    def process_message(self, msg):
        raise NotImplementedError

    def send_command(self, cmd_type, cmd_body: bytearray):
        cmd = MessageQuestCustom(
            self._device_type, self._protocol_version, cmd_type, cmd_body
        )
        try:
            self.build_send(cmd)
        except socket.error as e:
            _LOGGER.debug(
                f"[{self._device_id}] Interface send_command failure, {repr(e)}, "
                f"cmd_type: {cmd_type}, cmd_body: {cmd_body.hex()}"
            )

    def send_heartbeat(self):
        msg = PacketBuilder(self._device_id, bytearray([0x00])).finalize(msg_type=0)
        self.send_message(msg)

    def register_update(self, update):
        self._updates.append(update)

    def update_all(self, status):
        _LOGGER.debug(f"[{self._device_id}] Status update: {status}")
        for update in self._updates:
            update(status)

    def enable_device(self, available=True):
        self._available = available
        status = {"available": available}
        self.update_all(status)

    def open(self):
        if not self._is_run:
            self._is_run = True
            threading.Thread.start(self)

    def close(self):
        if self._is_run:
            self._is_run = False
            self.close_socket()

    def close_socket(self):
        self._unsupported_protocol = []
        self._buffer = b""
        if self._socket:
            self._socket.close()
            self._socket = None

    def set_ip_address(self, ip_address):
        if self._ip_address != ip_address:
            _LOGGER.debug(f"[{self._device_id}] Update IP address to {ip_address}")
            self._ip_address = ip_address
            self.close_socket()

    def set_refresh_interval(self, refresh_interval):
        self._refresh_interval = refresh_interval

    def run(self):
        while self._is_run:
            while self._socket is None:
                if self.connect(refresh_status=True) is False:
                    if not self._is_run:
                        return
                    self.close_socket()
                    time.sleep(5)
            timeout_counter = 0
            start = time.time()
            previous_refresh = start
            previous_heartbeat = start
            self._socket.settimeout(1)
            while True:
                try:
                    now = time.time()
                    if 0 < self._refresh_interval <= now - previous_refresh:
                        self.refresh_status()
                        previous_refresh = now
                    if now - previous_heartbeat >= self._heartbeat_interval:
                        self.send_heartbeat()
                        previous_heartbeat = now
                    msg = self._socket.recv(512)
                    msg_len = len(msg)
                    if msg_len == 0:
                        raise socket.error("Connection closed by peer")
                    result = self.parse_message(msg)
                    if result == ParseMessageResult.ERROR:
                        _LOGGER.debug(f"[{self._device_id}] Message 'ERROR' received")
                        self.close_socket()
                        break
                    elif result == ParseMessageResult.SUCCESS:
                        timeout_counter = 0
                except socket.timeout:
                    timeout_counter = timeout_counter + 1
                    if timeout_counter >= 120:
                        _LOGGER.debug(f"[{self._device_id}] Heartbeat timed out")
                        self.close_socket()
                        break
                except socket.error as e:
                    if self._is_run:
                        _LOGGER.debug(f"[{self._device_id}] Socket error {repr(e)}")
                        self.close_socket()
                    break
                except Exception as e:
                    _LOGGER.error(
                        f"[{self._device_id}] Unknown error :{e.__traceback__.tb_frame.f_globals['__file__']}, "
                        f"{e.__traceback__.tb_lineno}, {repr(e)}"
                    )
                    self.close_socket()
                    break

    def set_attribute(self, attr, value):
        raise NotImplementedError

    def get_attribute(self, attr):
        return self._attributes.get(attr)

    def set_customize(self, customize):
        pass

    @property
    def attributes(self):
        ret = {}
        for status in self._attributes.keys():
            ret[str(status)] = self._attributes[status]
        return ret



from ...message import MessageType, MessageRequest, MessageResponse, MessageBody


class MessageE8Base(MessageRequest):
    def __init__(self, protocol_version, message_type, body_type):
        super().__init__(
            device_type=0xE8,
            protocol_version=protocol_version,
            message_type=message_type,
            body_type=body_type,
        )

    @property
    def _body(self):
        raise NotImplementedError


class MessageQuery(MessageE8Base):
    def __init__(self, protocol_version):
        super().__init__(
            protocol_version=protocol_version,
            message_type=MessageType.query,
            body_type=0xAA,
        )

    @property
    def _body(self):
        return bytearray([0x55, 0x00, 0x01, 0x00, 0x00])


class E8MessageBody(MessageBody):
    def __init__(self, body):
        super().__init__(body)
        self.status = body[11]
        self.time_remaining = body[16] * 3600 + body[17] * 60 + body[18]
        self.keep_warm_remaining = body[19] * 3600 + body[20] * 60 + body[21]
        self.working_time = body[28] * 3600 + body[29] * 60 + body[30]
        self.target_temperature = body[39]
        self.current_temperature = body[39]
        self.finished = (body[41] & 0x01) > 0
        self.water_shortage = body[43] > 0


class MessageE8Response(MessageResponse):
    def __init__(self, message):
        super().__init__(message)
        if len(super().body) > 6:
            sub_cmd = super().body[6]
            if (
                (self.message_type == MessageType.set and sub_cmd in [0x02, 0x04, 0x06])
                or self.message_type in [MessageType.query, MessageType.notify1]
                and sub_cmd == 2
            ):
                self.set_body(E8MessageBody(super().body))
        self.set_attr()



import logging
from .message import MessageQuery, MessageFDResponse, MessageSet

try:
    from enum import StrEnum
except ImportError:
    from ...backports.enum import StrEnum
from ...device import MideaDevice

_LOGGER = logging.getLogger(__name__)


class DeviceAttributes(StrEnum):
    power = "power"
    fan_speed = "fan_speed"
    prompt_tone = "prompt_tone"
    target_humidity = "target_humidity"
    current_humidity = "current_humidity"
    current_temperature = "current_temperature"
    tank = "tank"
    mode = "mode"
    screen_display = "screen_display"
    disinfect = "disinfect"


class MideaFDDevice(MideaDevice):
    _modes = [
        "Manual",
        "Auto",
        "Continuous",
        "Living-Room",
        "Bed-Room",
        "Kitchen",
        "Sleep",
    ]
    _speeds_old = {
        1: "Lowest",
        40: "Low",
        60: "Medium",
        80: "High",
        102: "Auto",
        127: "Off",
    }
    _speeds_new = {
        1: "Lowest",
        39: "Low",
        59: "Medium",
        80: "High",
        101: "Auto",
        127: "Off",
    }
    _screen_displays = {0: "Bright", 6: "Dim", 7: "Off"}
    _detect_modes = ["Off", "PM 2.5", "Methanal"]

    def __init__(
        self,
        name: str,
        device_id: int,
        ip_address: str,
        port: int,
        token: str,
        key: str,
        protocol: int,
        model: str,
        subtype: int,
        customize: str,
    ):
        super().__init__(
            name=name,
            device_id=device_id,
            device_type=0xFD,
            ip_address=ip_address,
            port=port,
            token=token,
            key=key,
            protocol=protocol,
            model=model,
            subtype=subtype,
            attributes={
                DeviceAttributes.power: False,
                DeviceAttributes.fan_speed: None,
                DeviceAttributes.prompt_tone: True,
                DeviceAttributes.target_humidity: 60,
                DeviceAttributes.current_humidity: None,
                DeviceAttributes.current_temperature: None,
                DeviceAttributes.tank: 0,
                DeviceAttributes.mode: None,
                DeviceAttributes.screen_display: None,
                DeviceAttributes.disinfect: None,
            },
        )
        if self.subtype > 5:
            self._speeds = MideaFDDevice._speeds_new
        else:
            self._speeds = MideaFDDevice._speeds_old

    @property
    def modes(self):
        return list(MideaFDDevice._modes)

    @property
    def fan_speeds(self):
        return list(self._speeds.values())

    @property
    def screen_displays(self):
        return list(MideaFDDevice._screen_displays.values())

    @property
    def detect_modes(self):
        return self._detect_modes

    def build_query(self):
        return [MessageQuery(self._protocol_version)]

    def process_message(self, msg):
        message = MessageFDResponse(msg)
        _LOGGER.debug(f"[{self.device_id}] Received: {message}")
        new_status = {}
        for status in self._attributes.keys():
            if hasattr(message, str(status)):
                value = getattr(message, str(status))
                if status == DeviceAttributes.mode:
                    if value <= len(MideaFDDevice._modes):
                        self._attributes[status] = MideaFDDevice._modes[value - 1]
                    else:
                        self._attributes[status] = None
                elif status == DeviceAttributes.fan_speed:
                    if value in self._speeds.keys():
                        self._attributes[status] = self._speeds.get(value)
                    else:
                        self._attributes[status] = None
                elif status == DeviceAttributes.screen_display:
                    if value in MideaFDDevice._screen_displays.keys():
                        self._attributes[status] = MideaFDDevice._screen_displays.get(
                            value
                        )
                    else:
                        self._attributes[status] = None
                else:
                    self._attributes[status] = value
                new_status[str(status)] = self._attributes[status]
        return new_status

    def make_message_set(self):
        message = MessageSet(self._protocol_version)
        message.power = self._attributes[DeviceAttributes.power]
        message.prompt_tone = self._attributes[DeviceAttributes.prompt_tone]
        message.screen_display = self._attributes[DeviceAttributes.screen_display]
        message.disinfect = self._attributes[DeviceAttributes.disinfect]
        if self._attributes[DeviceAttributes.mode] in MideaFDDevice._modes:
            message.mode = (
                MideaFDDevice._modes.index(self._attributes[DeviceAttributes.mode]) + 1
            )
        else:
            message.mode = 1
        message.fan_speed = (
            40
            if self._attributes[DeviceAttributes.fan_speed] is None
            else list(self._speeds.keys())[
                list(self._speeds.values()).index(
                    self._attributes[DeviceAttributes.fan_speed]
                )
            ]
        )
        message.screen_display = (
            0
            if self._attributes[DeviceAttributes.screen_display] is None
            else list(MideaFDDevice._screen_displays.keys())[
                list(MideaFDDevice._screen_displays.values()).index(
                    self._attributes[DeviceAttributes.screen_display]
                )
            ]
        )
        return message

    def set_attribute(self, attr, value):
        if attr == DeviceAttributes.prompt_tone:
            self._attributes[DeviceAttributes.prompt_tone] = value
            self.update_all({DeviceAttributes.prompt_tone.value: value})
        else:
            message = self.make_message_set()
            if attr == DeviceAttributes.mode:
                if value in MideaFDDevice._modes:
                    message.mode = MideaFDDevice._modes.index(value) + 1
            elif attr == DeviceAttributes.fan_speed:
                if value in self._speeds.values():
                    message.fan_speed = list(self._speeds.keys())[
                        list(self._speeds.values()).index(value)
                    ]
            elif attr == DeviceAttributes.screen_display:
                if value in MideaFDDevice._screen_displays.values():
                    message.screen_display = list(
                        MideaFDDevice._screen_displays.keys()
                    )[list(MideaFDDevice._screen_displays.values()).index(value)]
                elif not value:
                    message.screen_display = 7
            else:
                setattr(message, str(attr), value)
            self.build_send(message)


class MideaAppliance(MideaFDDevice):
    pass



from ...crc8 import calculate
from ...message import (
    MessageType,
    MessageRequest,
    MessageResponse,
    MessageBody,
)


class MessageFCBase(MessageRequest):
    _message_serial = 0

    def __init__(self, protocol_version, message_type, body_type):
        super().__init__(
            device_type=0xFC,
            protocol_version=protocol_version,
            message_type=message_type,
            body_type=body_type,
        )
        MessageFCBase._message_serial += 1
        if MessageFCBase._message_serial >= 254:
            MessageFCBase._message_serial = 1
        self._message_id = MessageFCBase._message_serial

    @property
    def _body(self):
        raise NotImplementedError

    @property
    def body(self):
        body = bytearray([self.body_type]) + self._body + bytearray([self._message_id])
        body.append(calculate(body))
        return body


class MessageQuery(MessageFCBase):
    def __init__(self, protocol_version):
        super().__init__(
            protocol_version=protocol_version,
            message_type=MessageType.query,
            body_type=0x41,
        )

    @property
    def _body(self):
        return bytearray(
            [
                0x00,
                0x00,
                0xFF,
                0x03,
                0x00,
                0x00,
                0x02,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
            ]
        )


class MessageSet(MessageFCBase):
    def __init__(self, protocol_version):
        super().__init__(
            protocol_version=protocol_version,
            message_type=MessageType.set,
            body_type=0x48,
        )
        self.power = False
        self.mode = 0
        self.fan_speed = 0
        self.child_lock = False
        self.prompt_tone = False
        self.anion = False
        self.standby = False
        self.screen_display = 0
        self.detect_mode = 0
        self.standby_detect = [40, 20]

    @property
    def _body(self):
        # byte1 power
        power = 0x01 if self.power else 0x00
        detect = 0x08 if self.detect_mode > 0 else 0x00
        detect_mode = (self.detect_mode - 1) if self.detect_mode > 0 else 0
        # byte2 mode
        # byte3 fan_speed
        # byte 8 child_lock
        child_lock = 0x80 if self.child_lock else 0x00
        # byte 9 anion
        anion = 0x20 if self.anion else 0x00
        # byte 10 prompt_tone
        prompt_tone = 0x40 if self.prompt_tone else 0x00
        # byte 15/16/17 standby
        if self.standby:
            standby = 0x04
            standby_detect_high = self.standby_detect[0]
            standby_detect_low = self.standby_detect[1]
        else:
            standby = 0x08
            standby_detect_high = 0
            standby_detect_low = 0
        return bytearray(
            [
                power | prompt_tone | detect | 0x02,
                self.mode,
                self.fan_speed,
                0x00,
                0x00,
                0x00,
                0x00,
                child_lock,
                self.screen_display,
                anion,
                0x00,
                0x00,
                0x00,
                detect_mode,
                standby,
                standby_detect_high,
                standby_detect_low,
                0x00,
                0x00,
                0x00,
            ]
        )


class FCGeneralMessageBody(MessageBody):
    def __init__(self, body):
        super().__init__(body)
        self.power = (body[1] & 0x01) > 0
        self.mode = body[2] & 0xF0
        self.fan_speed = body[3] & 0x7F
        self.screen_display = body[9] & 0x07
        if len(body) > 14 and body[14] != 0xFF:
            self.pm25 = body[13] + (body[14] << 8)
        else:
            self.pm25 = None
        if len(body) > 15 and body[15] != 0xFF:
            self.tvoc = body[15]
        else:
            self.tvoc = None
        self.anion = (body[19] & 0x40 > 0) if len(body) > 19 else False
        self.standby = ((body[34] & 0xFF) == 0x14) if len(body) > 34 else False
        self.child_lock = (body[8] & 0x80 > 0) if len(body) > 8 else False
        if len(body) > 23:
            self.filter1_life = body[23]
        if len(body) > 24:
            self.filter2_life = body[24]
        if len(body) > 29:
            if (body[1] & 0x08) > 0:
                self.detect_mode = body[29] + 1
            else:
                self.detect_mode = 0
        if len(body) > 38 and body[38] != 0xFF:
            self.hcho = body[37] + (body[38] << 8)
        else:
            self.hcho = None


class FCNotifyMessageBody(MessageBody):
    def __init__(self, body):
        super().__init__(body)
        self.power = (body[1] & 0x01) > 0
        self.mode = body[2] & 0xF0
        self.fan_speed = body[3] & 0x7F
        self.screen_display = body[9] & 0x07
        if len(body) > 14 and body[14] != 0xFF:
            self.pm25 = body[13] + (body[14] << 8)
        else:
            self.pm25 = None
        if len(body) > 15 and body[15] != 0xFF:
            self.tvoc = body[15]
        else:
            self.tvoc = None
        self.anion = (body[10] & 0x20 > 0) if len(body) > 10 else False
        self.standby = (body[27] & 0x14 == 0xFF) if len(body) > 27 else False
        self.child_lock = (body[10] & 0x10 > 0) if len(body) > 10 else False
        if len(body) > 22:
            if (body[1] & 0x08) > 0:
                self.detect_mode = body[22] + 1
            else:
                self.detect_mode = 0
        if len(body) > 31 and body[31] != 0xFF:
            self.hcho = body[30] + (body[31] << 8)
        else:
            self.hcho = None


class MessageFCResponse(MessageResponse):
    def __init__(self, message):
        super().__init__(message)
        if self.body_type in [0xB0, 0xB1]:
            pass
        else:
            if (
                self.message_type
                in [MessageType.query, MessageType.set, MessageType.notify1]
                and self.body_type == 0xC8
            ):
                self.set_body(FCGeneralMessageBody(super().body))
            elif self.message_type == MessageType.notify1 and self.body_type == 0xA0:
                self.set_body(FCNotifyMessageBody(super().body))
        self.set_attr()



import logging
import math
from .message import MessageQuery, MessageSet, Message26Response

try:
    from enum import StrEnum
except ImportError:
    from ...backports.enum import StrEnum
from ...device import MideaDevice

_LOGGER = logging.getLogger(__name__)


class DeviceAttributes(StrEnum):
    main_light = "main_light"
    night_light = "night_light"
    mode = "mode"
    direction = "direction"
    current_humidity = "current_humidity"
    current_radar = "current_radar"
    current_temperature = "current_temperature"


class Midea26Device(MideaDevice):
    _modes = ["Off", "Heat(high)", "Heat(low)", "Bath", "Blow", "Ventilation", "Dry"]
    _directions = ["60", "70", "80", "90", "100", "110", "120", "Oscillate"]

    def __init__(
        self,
        name: str,
        device_id: int,
        ip_address: str,
        port: int,
        token: str,
        key: str,
        protocol: int,
        model: str,
        subtype: int,
        customize: str,
    ):
        super().__init__(
            name=name,
            device_id=device_id,
            device_type=0x26,
            ip_address=ip_address,
            port=port,
            token=token,
            key=key,
            protocol=protocol,
            model=model,
            subtype=subtype,
            attributes={
                DeviceAttributes.main_light: False,
                DeviceAttributes.night_light: False,
                DeviceAttributes.mode: None,
                DeviceAttributes.direction: None,
                DeviceAttributes.current_humidity: None,
                DeviceAttributes.current_radar: None,
                DeviceAttributes.current_temperature: None,
            },
        )
        self._fields = {}

    @staticmethod
    def _convert_to_midea_direction(direction):
        if direction == "Oscillate":
            result = 0xFD
        else:
            result = (
                Midea26Device._directions.index(direction) * 10 + 60
                if direction in Midea26Device._directions
                else 0xFD
            )
        return result

    @staticmethod
    def _convert_from_midea_direction(direction):
        if direction > 120 or direction < 60:
            result = 7
        else:
            result = math.floor((direction - 60 + 5) / 10)
        return result

    @property
    def preset_modes(self):
        return Midea26Device._modes

    @property
    def directions(self):
        return Midea26Device._directions

    def build_query(self):
        return [MessageQuery(self._protocol_version)]

    def process_message(self, msg):
        message = Message26Response(msg)
        _LOGGER.debug(f"[{self.device_id}] Received: {message}")
        new_status = {}
        self._fields = getattr(message, "fields")
        for status in self._attributes.keys():
            if hasattr(message, str(status)):
                value = getattr(message, str(status))
                if status == DeviceAttributes.mode:
                    self._attributes[status] = Midea26Device._modes[value]
                elif status == DeviceAttributes.direction:
                    self._attributes[status] = Midea26Device._directions[
                        self._convert_from_midea_direction(value)
                    ]
                else:
                    self._attributes[status] = value
                new_status[str(status)] = self._attributes[status]
        return new_status

    def set_attribute(self, attr, value):
        if attr in [
            DeviceAttributes.main_light,
            DeviceAttributes.night_light,
            DeviceAttributes.mode,
            DeviceAttributes.direction,
        ]:
            message = MessageSet(self._protocol_version)
            message.fields = self._fields
            message.main_light = self._attributes[DeviceAttributes.main_light]
            message.night_light = self._attributes[DeviceAttributes.night_light]
            message.mode = Midea26Device._modes.index(
                self._attributes[DeviceAttributes.mode]
            )
            message.direction = self._convert_to_midea_direction(
                self._attributes[DeviceAttributes.direction]
            )
            if attr in [DeviceAttributes.main_light, DeviceAttributes.night_light]:
                message.main_light = False
                message.night_light = False
                setattr(message, str(attr), value)
            elif attr == DeviceAttributes.mode:
                message.mode = Midea26Device._modes.index(value)
            elif attr == DeviceAttributes.direction:
                message.direction = self._convert_to_midea_direction(value)
            self.build_send(message)


class MideaAppliance(Midea26Device):
    pass



import logging
import json
from .message import MessageQuery, MessageSet, Message13Response

try:
    from enum import StrEnum
except ImportError:
    from ...backports.enum import StrEnum
from ...device import MideaDevice

_LOGGER = logging.getLogger(__name__)


class DeviceAttributes(StrEnum):
    brightness = "brightness"
    color_temperature = "color_temperature"
    rgb_color = "rgb_color"
    effect = "effect"
    power = "power"


class Midea13Device(MideaDevice):
    _effects = ["Manual", "Living", "Reading", "Mildly", "Cinema", "Night"]

    def __init__(
        self,
        name: str,
        device_id: int,
        ip_address: str,
        port: int,
        token: str,
        key: str,
        protocol: int,
        model: str,
        subtype: int,
        customize: str,
    ):
        super().__init__(
            name=name,
            device_id=device_id,
            device_type=0x13,
            ip_address=ip_address,
            port=port,
            token=token,
            key=key,
            protocol=protocol,
            model=model,
            subtype=subtype,
            attributes={
                DeviceAttributes.brightness: None,
                DeviceAttributes.color_temperature: None,
                DeviceAttributes.rgb_color: None,
                DeviceAttributes.effect: None,
                DeviceAttributes.power: False,
            },
        )
        self._color_temp_range = None
        self._default_color_temp_range = [2700, 6500]
        self.set_customize(customize)

    @property
    def effects(self):
        return Midea13Device._effects

    @property
    def color_temp_range(self):
        return self._color_temp_range

    def kelvin_to_midea(self, kelvin):
        return round(
            (kelvin - self._color_temp_range[0])
            / (self._color_temp_range[1] - self._color_temp_range[0])
            * 255
        )

    def midea_to_kelvin(self, midea):
        return (
            round((self._color_temp_range[1] - self._color_temp_range[0]) / 255 * midea)
            + self._color_temp_range[0]
        )

    def build_query(self):
        return [MessageQuery(self._protocol_version)]

    def process_message(self, msg):
        message = Message13Response(msg)
        _LOGGER.debug(f"[{self.device_id}] Received: {message}")
        new_status = {}
        if hasattr(message, "control_success"):
            new_status = {"control_success", message.control_success}
            if message.control_success:
                self.refresh_status()
        else:
            for status in self._attributes.keys():
                if hasattr(message, str(status)):
                    value = getattr(message, str(status))
                    if status == DeviceAttributes.effect:
                        self._attributes[status] = Midea13Device._effects[value]
                    elif status == DeviceAttributes.color_temperature:
                        self._attributes[status] = self.midea_to_kelvin(value)
                    else:
                        self._attributes[status] = value
                    new_status[str(status)] = self._attributes[status]
        return new_status

    def set_attribute(self, attr, value):
        if attr in [
            DeviceAttributes.brightness,
            DeviceAttributes.color_temperature,
            DeviceAttributes.effect,
            DeviceAttributes.power,
        ]:
            message = MessageSet(self._protocol_version)
            if attr == DeviceAttributes.effect and value in self._effects:
                setattr(message, str(attr), Midea13Device._effects.index(value))
            elif attr == DeviceAttributes.color_temperature:
                setattr(message, str(attr), self.kelvin_to_midea(value))
            else:
                setattr(message, str(attr), value)
            self.build_send(message)

    def set_customize(self, customize):
        self._color_temp_range = self._default_color_temp_range
        if customize and len(customize) > 0:
            try:
                params = json.loads(customize)
                if params and "color_temp_range_kelvin" in params:
                    self._color_temp_range = params.get("color_temp_range_kelvin")
            except Exception as e:
                _LOGGER.error(f"[{self.device_id}] Set customize error: {repr(e)}")
            self.update_all({"color_temp_range": self._color_temp_range})


class MideaAppliance(Midea13Device):
    pass



import hmac
from hashlib import md5, sha256
from typing import Any
from urllib.parse import unquote_plus, urlencode, urlparse

from Crypto.Cipher import AES
from Crypto.Random import get_random_bytes
from Crypto.Util.Padding import pad, unpad
from Crypto.Util.strxor import strxor

MSGTYPE_HANDSHAKE_REQUEST = 0x0
MSGTYPE_HANDSHAKE_RESPONSE = 0x1
MSGTYPE_ENCRYPTED_RESPONSE = 0x3
MSGTYPE_ENCRYPTED_REQUEST = 0x6


class CloudSecurity:
    def __init__(self, login_key, iot_key, hmac_key, fixed_key=None, fixed_iv=None):
        self._login_key = login_key
        self._iot_key = iot_key
        self._hmac_key = hmac_key
        self._aes_key = None
        self._aes_iv = None
        self._fixed_key = format(fixed_key, "x").encode("ascii") if fixed_key else None
        self._fixed_iv = format(fixed_iv, "x").encode("ascii") if fixed_iv else None

    def sign(self, url: str, data: Any, random: str) -> str:
        msg = self._iot_key
        msg += str(data)
        msg += random
        sign = hmac.new(self._hmac_key.encode("ascii"), msg.encode("ascii"), sha256)
        return sign.hexdigest()

    def encrypt_password(self, login_id, data):
        m = sha256()
        m.update(data.encode("ascii"))
        login_hash = login_id + m.hexdigest() + self._login_key
        m = sha256()
        m.update(login_hash.encode("ascii"))
        return m.hexdigest()

    def encrypt_iam_password(self, login_id, data) -> str:
        raise NotImplementedError

    @staticmethod
    def get_deviceid(username):
        return sha256(f"Hello, {username}!".encode("ascii")).digest().hex()[:16]

    @staticmethod
    def get_udp_id(appliance_id, method=0):
        if method == 0:
            bytes_id = bytes(reversed(appliance_id.to_bytes(8, "big")))
        elif method == 1:
            bytes_id = appliance_id.to_bytes(6, "big")
        elif method == 2:
            bytes_id = appliance_id.to_bytes(6, "little")
        else:
            return None
        data = bytearray(sha256(bytes_id).digest())
        for i in range(0, 16):
            data[i] ^= data[i + 16]
        return data[0:16].hex()

    def set_aes_keys(self, key, iv):
        if isinstance(key, str):
            key = key.encode("ascii")
        if isinstance(iv, str):
            iv = iv.encode("ascii")
        self._aes_key = key
        self._aes_iv = iv

    def aes_encrypt_with_fixed_key(self, data):
        return self.aes_encrypt(data, self._fixed_key, self._fixed_iv)

    def aes_decrypt_with_fixed_key(self, data):
        return self.aes_decrypt(data, self._fixed_key, self._fixed_iv)

    def aes_encrypt(self, data, key=None, iv=None):
        if key is not None:
            aes_key = key
            aes_iv = iv
        else:
            aes_key = self._aes_key
            aes_iv = self._aes_iv
        if aes_key is None:
            raise ValueError("Encrypt need a key")
        if isinstance(data, str):
            data = bytes.fromhex(data)
        if aes_iv is None:  # ECB
            return AES.new(aes_key, AES.MODE_ECB).encrypt(pad(data, 16))
        else:  # CBC
            return AES.new(aes_key, AES.MODE_CBC, iv=aes_iv).encrypt(pad(data, 16))

    def aes_decrypt(self, data, key=None, iv=None):
        if key is not None:
            aes_key = key
            aes_iv = iv
        else:
            aes_key = self._aes_key
            aes_iv = self._aes_iv
        if aes_key is None:
            raise ValueError("Encrypt need a key")
        if isinstance(data, str):
            data = bytes.fromhex(data)
        if aes_iv is None:  # ECB
            return unpad(
                AES.new(aes_key, AES.MODE_ECB).decrypt(data), len(aes_key)
            ).decode()
        else:  # CBC
            return unpad(
                AES.new(aes_key, AES.MODE_CBC, iv=aes_iv).decrypt(data), len(aes_key)
            ).decode()


class MeijuCloudSecurity(CloudSecurity):
    def __init__(self, login_key, iot_key, hmac_key):
        super().__init__(login_key, iot_key, hmac_key, 10864842703515613082)

    def encrypt_iam_password(self, login_id, data) -> str:
        md = md5()
        md.update(data.encode("ascii"))
        md_second = md5()
        md_second.update(md.hexdigest().encode("ascii"))
        return md_second.hexdigest()


class MSmartCloudSecurity(CloudSecurity):
    def __init__(self, login_key, iot_key, hmac_key):
        super().__init__(
            login_key, iot_key, hmac_key, 13101328926877700970, 16429062708050928556
        )

    def encrypt_iam_password(self, login_id, data) -> str:
        md = md5()
        md.update(data.encode("ascii"))
        md_second = md5()
        md_second.update(md.hexdigest().encode("ascii"))
        login_hash = login_id + md_second.hexdigest() + self._login_key
        sha = sha256()
        sha.update(login_hash.encode("ascii"))
        return sha.hexdigest()

    def set_aes_keys(self, encrypted_key, encrypted_iv):
        key_digest = sha256(self._login_key.encode("ascii")).hexdigest()
        tmp_key = key_digest[:16].encode("ascii")
        tmp_iv = key_digest[16:32].encode("ascii")
        self._aes_key = self.aes_decrypt(encrypted_key, tmp_key, tmp_iv).encode("ascii")
        self._aes_iv = self.aes_decrypt(encrypted_iv, tmp_key, tmp_iv).encode("ascii")


class MideaAirSecurity(CloudSecurity):
    def __init__(self, login_key):
        super().__init__(login_key, None, None)

    def sign(self, url: str, data: Any, random: str) -> str:
        payload = unquote_plus(urlencode(sorted(data.items(), key=lambda x: x[0])))
        sha = sha256()
        sha.update((urlparse(url).path + payload + self._login_key).encode("ascii"))
        return sha.hexdigest()


class LocalSecurity:
    def __init__(self):
        self.blockSize = 16
        self.iv = b"\0" * 16
        self.aes_key = bytes.fromhex(
            format(141661095494369103254425781617665632877, "x")
        )
        self.salt = bytes.fromhex(
            format(
                233912452794221312800602098970898185176935770387238278451789080441632479840061417076563,
                "x",
            )
        )
        self._tcp_key = None
        self._request_count = 0
        self._response_count = 0

    def aes_decrypt(self, raw):
        try:
            return unpad(
                AES.new(self.aes_key, AES.MODE_ECB).decrypt(bytearray(raw)), 16
            )
        except ValueError:
            return bytearray(0)

    def aes_encrypt(self, raw):
        return AES.new(self.aes_key, AES.MODE_ECB).encrypt(bytearray(pad(raw, 16)))

    def aes_cbc_decrypt(self, raw, key):
        return AES.new(key=key, mode=AES.MODE_CBC, iv=self.iv).decrypt(raw)

    def aes_cbc_encrypt(self, raw, key):
        return AES.new(key=key, mode=AES.MODE_CBC, iv=self.iv).encrypt(raw)

    def encode32_data(self, raw):
        return md5(raw + self.salt).digest()

    def tcp_key(self, response, key):
        if response == b"ERROR":
            raise Exception("authentication failed")
        if len(response) != 64:
            raise Exception("unexpected data length")
        payload = response[:32]
        sign = response[32:]
        plain = self.aes_cbc_decrypt(payload, key)
        if sha256(plain).digest() != sign:
            raise Exception("sign does not match")
        self._tcp_key = strxor(plain, key)
        self._request_count = 0
        self._response_count = 0
        return self._tcp_key

    def encode_8370(self, data, msgtype):
        header = bytearray([0x83, 0x70])
        size, padding = len(data), 0
        if msgtype in (MSGTYPE_ENCRYPTED_RESPONSE, MSGTYPE_ENCRYPTED_REQUEST):
            if (size + 2) % 16 != 0:
                padding = 16 - (size + 2 & 0xF)
                size += padding + 32
                data += get_random_bytes(padding)
        header += size.to_bytes(2, "big")
        header += bytearray([0x20, padding << 4 | msgtype])
        data = self._request_count.to_bytes(2, "big") + data
        self._request_count += 1
        if self._request_count >= 0xFFFF:
            self._request_count = 0
        if msgtype in (MSGTYPE_ENCRYPTED_RESPONSE, MSGTYPE_ENCRYPTED_REQUEST):
            sign = sha256(header + data).digest()
            data = self.aes_cbc_encrypt(raw=data, key=self._tcp_key) + sign
        return header + data

    def decode_8370(self, data):
        if len(data) < 6:
            return [], data
        header = data[:6]
        if header[0] != 0x83 or header[1] != 0x70:
            raise Exception("not an 8370 message")
        size = int.from_bytes(header[2:4], "big") + 8
        leftover = None
        if len(data) < size:
            return [], data
        elif len(data) > size:
            leftover = data[size:]
            data = data[:size]
        if header[4] != 0x20:
            raise Exception("missing byte 4")
        padding = header[5] >> 4
        msgtype = header[5] & 0xF
        data = data[6:]
        if msgtype in (MSGTYPE_ENCRYPTED_RESPONSE, MSGTYPE_ENCRYPTED_REQUEST):
            sign = data[-32:]
            data = data[:-32]
            data = self.aes_cbc_decrypt(raw=data, key=self._tcp_key)
            if sha256(header + data).digest() != sign:
                raise Exception("sign does not match")
            if padding:
                data = data[:-padding]
        self._response_count = int.from_bytes(data[:2], "big")
        data = data[2:]
        if leftover:
            packets, incomplete = self.decode_8370(leftover)
            return [data] + packets, incomplete
        return [data], b""



from ...message import (
    MessageType,
    MessageRequest,
    MessageResponse,
    MessageBody,
)


class MessageECBase(MessageRequest):
    def __init__(self, protocol_version, message_type, body_type):
        super().__init__(
            device_type=0xEC,
            protocol_version=protocol_version,
            message_type=message_type,
            body_type=body_type,
        )

    @property
    def _body(self):
        raise NotImplementedError


class MessageQuery(MessageECBase):
    def __init__(self, protocol_version):
        super().__init__(
            protocol_version=protocol_version,
            message_type=MessageType.query,
            body_type=None,
        )

    @property
    def body(self):
        return bytearray([0xAA, 0x55, 0x01, 0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00])

    @property
    def _body(self):
        return bytearray([])


class ECGeneralMessageBody(MessageBody):
    def __init__(self, body):
        super().__init__(body)
        self.mode = body[4] + (body[5] << 8)
        self.progress = body[8]
        self.cooking = self.progress == 1
        self.time_remaining = body[12] * 60 + body[13]
        self.keep_warm_time = body[16] * 60 + body[17]
        self.top_temperature = body[21]
        self.bottom_temperature = body[22]
        self.with_pressure = (body[23] & 0x04) > 0


class ECBodyNew(MessageBody):
    def __init__(self, body):
        super().__init__(body)
        self.progress = body[11]
        self.cooking = self.progress == 1
        self.time_remaining = body[16] * 60 + body[17]
        self.keep_warm_time = body[19] * 60 + body[20]
        self.top_temperature = body[48]
        self.bottom_temperature = body[49]
        self.with_pressure = body[33] > 0


class MessageECResponse(MessageResponse):
    def __init__(self, message):
        super().__init__(message)
        if self.message_type == MessageType.notify1 and super().body[3] == 0x01:
            self.set_body(ECBodyNew(super().body))
        elif (
            (self.message_type == MessageType.set and super().body[3] == 0x02)
            or (self.message_type == MessageType.query and super().body[3] == 0x03)
            or (self.message_type == MessageType.notify1 and super().body[3] == 0x04)
            or (self.message_type == MessageType.notify1 and super().body[3] == 0x3D)
        ):
            self.set_body(ECGeneralMessageBody(super().body))
        elif self.message_type == MessageType.notify1 and super().body[3] == 0x06:
            self.mode = super().body[4] + (super().body[5] << 8)
        self.set_attr()



from ...message import MessageType, MessageRequest, MessageResponse, MessageBody


class MessageB3Base(MessageRequest):
    def __init__(self, protocol_version, message_type, body_type):
        super().__init__(
            device_type=0xB3,
            protocol_version=protocol_version,
            message_type=message_type,
            body_type=body_type,
        )

    @property
    def _body(self):
        raise NotImplementedError


class MessageQuery(MessageB3Base):
    def __init__(self, protocol_version):
        super().__init__(
            protocol_version=protocol_version,
            message_type=MessageType.query,
            body_type=0x31,
        )

    @property
    def _body(self):
        return bytearray([])


class B3MessageBody31(MessageBody):
    def __init__(self, body):
        super().__init__(body)
        self.top_compartment_status = body[1]
        self.top_compartment_mode = body[2]
        self.top_compartment_temperature = body[3]
        self.top_compartment_remaining = (
            body[23] * 3600
            if len(body) > 23 and body[23] != 0xFF
            else (
                0 + body[4] * 60
                if body[4] != 0xFF
                else 0 + body[5]
                if body[5] != 0xFF
                else 0
            )
        )
        self.bottom_compartment_status = body[6]
        self.bottom_compartment_mode = body[7]
        self.bottom_compartment_temperature = body[8]
        self.bottom_compartment_remaining = (
            body[24] * 3600
            if len(body) > 24 and body[24] != 0xFF
            else (
                0 + body[9] * 60
                if body[9] != 0xFF
                else 0 + body[10]
                if body[10] != 0xFF
                else 0
            )
        )
        self.middle_compartment_status = body[17]
        self.middle_compartment_mode = body[18]
        self.middle_compartment_temperature = body[19]
        self.middle_compartment_remaining = (
            body[25] * 3600
            if len(body) > 25 and body[25] != 0xFF
            else (
                0 + body[20] * 60
                if body[20] != 0xFF
                else 0 + body[21]
                if body[21] != 0xFF
                else 0
            )
        )
        self.lock = body[11] & 0x01 > 0
        self.bottom_compartment_door = body[11] & 0x02 > 0
        self.top_compartment_door = body[11] & 0x04 > 0
        self.middle_compartment_door = body[11] & 0x10 > 0
        self.bottom_compartment_preheating = body[16] & 0x01 > 0
        self.top_compartment_preheating = body[16] & 0x02 > 0
        self.middle_compartment_preheating = body[16] & 0x10 > 0
        self.bottom_compartment_cooling = body[16] & 0x04 > 0
        self.top_compartment_cooling = body[16] & 0x08 > 0
        self.middle_compartment_cooling = body[16] & 0x20 > 0


class B3MessageBody21(MessageBody):
    def __init__(self, body):
        super().__init__(body)
        self.top_compartment_status = body[1]
        self.top_compartment_mode = body[2]
        self.top_compartment_temperature = body[3]
        self.top_compartment_remaining = (
            body[17] * 3600
            if len(body) > 17 and body[17] != 0xFF
            else (
                0 + body[4] * 60
                if body[4] != 0xFF
                else 0 + body[5]
                if body[5] != 0xFF
                else 0
            )
        )
        self.bottom_compartment_status = body[6]
        self.bottom_compartment_mode = body[7]
        self.bottom_compartment_temperature = body[8]
        self.bottom_compartment_remaining = (
            body[18] * 3600
            if len(body) > 18 and body[18] != 0xFF
            else (
                0 + body[9] * 60
                if body[9] != 0xFF
                else 0 + body[10]
                if body[10] != 0xFF
                else 0
            )
        )
        self.middle_compartment_status = body[12]
        self.middle_compartment_mode = body[13]
        self.middle_compartment_temperature = body[14]
        self.middle_compartment_remaining = (
            body[19] * 3600
            if len(body) > 19 and body[19] != 0xFF
            else (
                0 + body[15] * 60
                if body[15] != 0xFF
                else 0 + body[16]
                if body[16] != 0xFF
                else 0
            )
        )
        self.lock = body[11] & 0x01 > 0


class B3MessageBody24(MessageBody):
    def __init__(self, body):
        super().__init__(body)
        self.top_compartment_status = body[5]
        self.top_compartment_mode = body[6]
        self.top_compartment_temperature = body[7]
        self.top_compartment_remaining = (
            body[8] * 60 if body[8] != 0xFF else 0 + body[9] if body[9] != 0xFF else 0
        )
        self.bottom_compartment_status = body[10]
        self.bottom_compartment_mode = body[11]
        self.bottom_compartment_temperature = body[12]
        self.bottom_compartment_remaining = (
            body[13] * 60
            if body[13] != 0xFF
            else 0 + body[14]
            if body[14] != 0xFF
            else 0
        )
        self.bottom_compartment_status = body[15]
        self.bottom_compartment_mode = body[16]
        self.bottom_compartment_temperature = body[17]
        self.bottom_compartment_remaining = (
            body[18] * 60
            if body[18] != 0xFF
            else 0 + body[19]
            if body[19] != 0xFF
            else 0
        )


class MessageB3Response(MessageResponse):
    def __init__(self, message):
        super().__init__(message)
        if (
            self.message_type == MessageType.query
            and self.body_type == 0x31
            or self.message_type == MessageType.notify1
            and self.body_type == 0x41
        ):
            self.set_body(B3MessageBody31(super().body))
        elif self.message_type == MessageType.set and self.body_type == 0x21:
            self.set_body(B3MessageBody21(super().body))
        elif self.message_type == MessageType.set and self.body_type == 0x24:
            self.set_body(B3MessageBody21(super().body))
        self.set_attr()



from ...message import (
    MessageType,
    MessageRequest,
    MessageResponse,
    MessageBody,
)


class MessageFABase(MessageRequest):
    def __init__(self, protocol_version, message_type, body_type):
        super().__init__(
            device_type=0xCE,
            protocol_version=protocol_version,
            message_type=message_type,
            body_type=body_type,
        )

    @property
    def _body(self):
        raise NotImplementedError


class MessageQuery(MessageFABase):
    def __init__(self, protocol_version):
        super().__init__(
            protocol_version=protocol_version,
            message_type=MessageType.query,
            body_type=0x01,
        )

    @property
    def _body(self):
        return bytearray([])


class MessageSet(MessageFABase):
    def __init__(self, protocol_version):
        super().__init__(
            protocol_version=protocol_version,
            message_type=MessageType.set,
            body_type=0x01,
        )

        self.power = False
        self.fan_speed = 0
        self.link_to_ac = False
        self.sleep_mode = False
        self.eco_mode = False
        self.aux_heating = False
        self.powerful_purify = False
        self.scheduled = False
        self.child_lock = False

    @property
    def _body(self):
        power = 0x80 if self.power else 0x00
        link_to_ac = 0x01 if self.link_to_ac else 0x00
        sleep_mode = 0x02 if self.sleep_mode else 0x00
        eco_mode = 0x04 if self.eco_mode else 0x00
        aux_heating = 0x08 if self.aux_heating else 0x00
        powerful_purify = 0x10 if self.powerful_purify else 0x00
        scheduled = 0x01 if self.scheduled else 0x00
        child_lock = 0x7F if self.child_lock else 0x00
        return bytearray(
            [
                power | 0x01,
                self.fan_speed,
                link_to_ac | sleep_mode | eco_mode | aux_heating | powerful_purify,
                scheduled,
                0x00,
                child_lock,
            ]
        )


class CEGeneralMessageBody(MessageBody):
    def __init__(self, body):
        super().__init__(body)
        self.power = (body[1] & 0x80) > 0
        self.child_lock = (body[1] & 0x20) > 0
        self.scheduled = (body[1] & 0x40) > 0
        self.fan_speed = body[2]
        self.pm25 = (body[3] << 8) + body[4]
        self.co2 = (body[5] << 8) + body[6]
        if body[7] != 0xFF:
            self.current_humidity = (body[7] << 8) + body[8] / 10
        else:
            self.current_humidity = None
        if body[9] != 0xFF:
            self.current_temperature = (body[9] << 8) + (body[10] - 60) / 2
        else:
            self.current_temperature = None
        if body[11] != 0xFF:
            self.hcho = (body[11] << 8) + body[12] / 1000
        else:
            self.hcho = None
        self.link_to_ac = (body[17] & 0x01) > 0
        self.sleep_mode = (body[17] & 0x02) > 0
        self.eco_mode = (body[17] & 0x04) > 0
        if (body[19] & 0x02) > 0:
            self.aux_heating = (body[17] & 0x08) > 0
        else:
            self.aux_heating = None
        self.powerful_purify = (body[17] & 0x10) > 0
        self.filter_cleaning_reminder = (body[18] & 0x01) > 0
        self.filter_change_reminder = (body[18] & 0x02) > 0
        self.error_code = body[24]


class CENotifyMessageBody(MessageBody):
    def __init__(self, body):
        super().__init__(body)
        self.pm25 = (body[1] << 8) + body[2]
        self.co2 = (body[3] << 8) + body[4]
        if body[5] != 0xFF:
            self.current_humidity = (body[5] << 8) + body[6] / 10
        else:
            self.current_humidity = None
        if body[7] != 0xFF:
            self.current_temperature = (body[7] << 8) + (body[8] - 60) / 2
        else:
            self.current_temperature = None
        if body[9] != 0xFF:
            self.hcho = (body[9] << 8) + body[10] / 1000
        else:
            self.hcho = None
        self.error_code = body[12]


class MessageCEResponse(MessageResponse):
    def __init__(self, message):
        super().__init__(message)
        if (
            self.message_type in [MessageType.query, MessageType.set]
            and self.body_type == 0x01
        ) or (self.message_type == MessageType.notify1 and self.body_type == 0x02):
            self.set_body(CEGeneralMessageBody(super().body))
        elif self.message_type == MessageType.notify1 and self.body_type == 0x01:
            self.set_body(CENotifyMessageBody(super().body))
        self.set_attr()



import logging
from abc import ABC
from enum import IntEnum

_LOGGER = logging.getLogger(__name__)


class MessageLenError(Exception):
    pass


class MessageBodyError(Exception):
    pass


class MessageCheckSumError(Exception):
    pass


class MessageType(IntEnum):
    set = (0x02,)
    query = (0x03,)
    notify1 = (0x04,)
    notify2 = (0x05,)
    exception = (0x06,)
    exception2 = (0x0A,)
    query_appliance = 0xA0


class MessageBase(ABC):
    HEADER_LENGTH = 10

    def __init__(self):
        self._device_type = 0x00
        self._message_type = 0x00
        self._body_type = 0x00
        self._protocol_version = 0x00

    @staticmethod
    def checksum(data):
        return (~sum(data) + 1) & 0xFF

    @property
    def header(self):
        raise NotImplementedError

    @property
    def body(self):
        raise NotImplementedError

    @property
    def message_type(self):
        return self._message_type

    @message_type.setter
    def message_type(self, value):
        self._message_type = value

    @property
    def device_type(self):
        return self._device_type

    @device_type.setter
    def device_type(self, value):
        self._device_type = value

    @property
    def body_type(self):
        return self._body_type

    @body_type.setter
    def body_type(self, value):
        self._body_type = value

    @property
    def protocol_version(self):
        return self._protocol_version

    @protocol_version.setter
    def protocol_version(self, protocol_version):
        self._protocol_version = protocol_version

    def __str__(self) -> str:
        output = {
            "header": self.header.hex(),
            "body": self.body.hex(),
            "message type": "%02x" % self._message_type,
            "body type": (
                ("%02x" % self._body_type) if self._body_type is not None else "None"
            ),
        }
        return str(output)


class MessageRequest(MessageBase):
    def __init__(self, device_type, protocol_version, message_type, body_type):
        super().__init__()
        self.device_type = device_type
        self.protocol_version = protocol_version
        self.message_type = message_type
        self.body_type = body_type

    @property
    def header(self):
        length = self.HEADER_LENGTH + len(self.body)
        return bytearray(
            [
                # flag
                0xAA,
                # length
                length,
                # device type
                self.device_type,
                # frame checksum
                0x00,  # self._device_type ^ length,
                # unused
                0x00,
                0x00,
                # frame ID
                0x00,
                # frame protocol version
                0x00,
                # device protocol version
                self.protocol_version,
                # frame type
                self.message_type,
            ]
        )

    @property
    def _body(self):
        raise NotImplementedError

    @property
    def body(self):
        body = bytearray([])
        if self.body_type is not None:
            body.append(self.body_type)
        if self._body is not None:
            body.extend(self._body)
        return body

    def serialize(self):
        stream = self.header + self.body
        stream.append(MessageBase.checksum(stream[1:]))
        return stream


class MessageQuestCustom(MessageRequest):
    def __init__(self, device_type, protocol_version, cmd_type, cmd_body):
        super().__init__(
            device_type=device_type,
            protocol_version=protocol_version,
            message_type=cmd_type,
            body_type=None,
        )
        self._cmd_body = cmd_body

    @property
    def _body(self):
        return bytearray([])

    @property
    def body(self):
        return self._cmd_body


class MessageQueryAppliance(MessageRequest):
    def __init__(self, device_type):
        super().__init__(
            device_type=device_type,
            protocol_version=0,
            message_type=MessageType.query_appliance,
            body_type=None,
        )

    @property
    def _body(self):
        return bytearray([])

    @property
    def body(self):
        return bytearray([0x00] * 19)


class MessageBody:
    def __init__(self, body):
        self._data = body

    @property
    def data(self):
        return self._data

    @property
    def body_type(self):
        return self._data[0]

    @staticmethod
    def read_byte(body, byte, default_value=0):
        return body[byte] if len(body) > byte else default_value


class NewProtocolMessageBody(MessageBody):
    def __init__(self, body, bt):
        super().__init__(body)
        if bt == 0xB5:
            self._pack_len = 4
        else:
            self._pack_len = 5

    @staticmethod
    def pack(param, value: bytearray, pack_len=4):
        length = len(value)
        if pack_len == 4:
            stream = bytearray([param & 0xFF, param >> 8, length]) + value
        else:
            stream = bytearray([param & 0xFF, param >> 8, 0x00, length]) + value
        return stream

    def parse(self):
        result = {}
        try:
            pos = 2
            for pack in range(0, self.data[1]):
                param = self.data[pos] + (self.data[pos + 1] << 8)
                if self._pack_len == 5:
                    pos += 1
                length = self.data[pos + 2]
                if length > 0:
                    value = self.data[pos + 3 : pos + 3 + length]
                    result[param] = value
                pos += 3 + length
        except IndexError:
            # Some device used non-standard new-protocol(美的乐享三代中央空调?)
            _LOGGER.debug(f"Non-standard new-protocol {self.data.hex()}")
        return result


class MessageResponse(MessageBase):
    def __init__(self, message):
        super().__init__()
        if message is None or len(message) < self.HEADER_LENGTH + 1:
            raise MessageLenError
        self._header = message[: self.HEADER_LENGTH]
        self.protocol_version = self._header[-2]
        self.message_type = self._header[-1]
        self.device_type = self._header[2]
        body = message[self.HEADER_LENGTH : -1]
        self._body = MessageBody(body)
        self.body_type = self._body.body_type

    @property
    def header(self):
        return self._header

    @property
    def body(self):
        return self._body.data

    def set_body(self, body: MessageBody):
        self._body = body

    def set_attr(self):
        for key in vars(self._body).keys():
            if key != "data":
                value = getattr(self._body, key, None)
                setattr(self, key, value)


class MessageApplianceResponse(MessageResponse):
    def __init__(self, message):
        super().__init__(message)



#!/bin/python
"""setup midea-local."""

import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

requires = ["aiohttp", "ifaddr", "pycryptodome"]

setuptools.setup(
    name="midea-local",
    version="1.0.2",
    author="rokam",
    author_email="lucas@mindello.com.br",
    description="Control your Midea M-Smart appliances via local area network",
    license="MIT",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/rokam/midea-local",
    install_requires=requires,
    packages=setuptools.find_packages(
        include=["midealocal", "midealocal.*"], exclude=["tests", "tests.*"]
    ),
    python_requires=">=3.10",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



import logging
import json
from .message import MessageQuery, MessageFAResponse, MessageSet

try:
    from enum import StrEnum
except ImportError:
    from ...backports.enum import StrEnum
from ...device import MideaDevice

_LOGGER = logging.getLogger(__name__)


class DeviceAttributes(StrEnum):
    power = "power"
    child_lock = "child_lock"
    mode = "mode"
    fan_speed = "fan_speed"
    oscillate = "oscillate"
    oscillation_angle = "oscillation_angle"
    tilting_angle = "tilting_angle"
    oscillation_mode = "oscillation_mode"


class MideaFADevice(MideaDevice):
    _oscillation_angles = ["Off", "30", "60", "90", "120", "180", "360"]
    _tilting_angles = ["Off", "30", "60", "90", "120", "180", "360", "+60", "-60", "40"]
    _oscillation_modes = [
        "Off",
        "Oscillation",
        "Tilting",
        "Curve-W",
        "Curve-8",
        "Reserved",
        "Both",
    ]
    _modes = [
        "Normal",
        "Natural",
        "Sleep",
        "Comfort",
        "Silent",
        "Baby",
        "Induction",
        "Circulation",
        "Strong",
        "Soft",
        "Customize",
        "Warm",
        "Smart",
    ]

    def __init__(
        self,
        name: str,
        device_id: int,
        ip_address: str,
        port: int,
        token: str,
        key: str,
        protocol: int,
        model: str,
        subtype: int,
        customize: str,
    ):
        super().__init__(
            name=name,
            device_id=device_id,
            device_type=0xFA,
            ip_address=ip_address,
            port=port,
            token=token,
            key=key,
            protocol=protocol,
            model=model,
            subtype=subtype,
            attributes={
                DeviceAttributes.power: False,
                DeviceAttributes.child_lock: False,
                DeviceAttributes.mode: 0,
                DeviceAttributes.fan_speed: 0,
                DeviceAttributes.oscillate: False,
                DeviceAttributes.oscillation_angle: None,
                DeviceAttributes.tilting_angle: None,
                DeviceAttributes.oscillation_mode: None,
            },
        )
        self._default_speed_count = 3
        self._speed_count = self._default_speed_count
        self.set_customize(customize)

    @property
    def speed_count(self):
        return self._speed_count

    @property
    def oscillation_angles(self):
        return MideaFADevice._oscillation_angles

    @property
    def tilting_angles(self):
        return MideaFADevice._tilting_angles

    @property
    def oscillation_modes(self):
        return MideaFADevice._oscillation_modes

    @property
    def preset_modes(self):
        return self._modes

    def build_query(self):
        return [MessageQuery(self._protocol_version)]

    def process_message(self, msg):
        message = MessageFAResponse(msg)
        _LOGGER.debug(f"[{self.device_id}] Received: {message}")
        new_status = {}
        for status in self._attributes.keys():
            if hasattr(message, str(status)):
                value = getattr(message, str(status))
                if status == DeviceAttributes.oscillation_angle:
                    if value < len(MideaFADevice._oscillation_angles):
                        self._attributes[status] = MideaFADevice._oscillation_angles[
                            value
                        ]
                    else:
                        self._attributes[status] = None
                elif status == DeviceAttributes.tilting_angle:
                    if value < len(MideaFADevice._tilting_angles):
                        self._attributes[status] = MideaFADevice._tilting_angles[value]
                    else:
                        self._attributes[status] = None
                elif status == DeviceAttributes.oscillation_mode:
                    if value < len(MideaFADevice._oscillation_modes):
                        self._attributes[status] = MideaFADevice._oscillation_modes[
                            value
                        ]
                    else:
                        self._attributes[status] = None
                elif status == DeviceAttributes.mode:
                    if value < len(MideaFADevice._modes):
                        self._attributes[status] = MideaFADevice._modes[value]
                    else:
                        self._attributes[status] = None
                elif status == DeviceAttributes.power:
                    self._attributes[status] = value
                    if not value:
                        self._attributes[DeviceAttributes.fan_speed] = 0
                elif (
                    status == DeviceAttributes.fan_speed
                    and not self._attributes[DeviceAttributes.power]
                ):
                    self._attributes[status] = 0
                else:
                    self._attributes[status] = value
                new_status[str(status)] = self._attributes[status]
        return new_status

    def set_oscillation(self, attr, value):
        message = None
        if self._attributes[attr] != value:
            if attr == DeviceAttributes.oscillate:
                message = MessageSet(self._protocol_version, self.subtype)
                message.oscillate = value
                if value:
                    message.oscillation_angle = 3  # 90
                    message.oscillation_mode = 1  # Oscillation
            elif attr == DeviceAttributes.oscillation_mode and (
                value in MideaFADevice._oscillation_modes or not value
            ):
                message = MessageSet(self._protocol_version, self.subtype)
                if value == "Off" or not value:
                    message.oscillate = False
                else:
                    message.oscillate = True
                    message.oscillation_mode = MideaFADevice._oscillation_modes.index(
                        value
                    )
                    if value == "Oscillation":
                        if (
                            self._attributes[DeviceAttributes.oscillation_angle]
                            == "Off"
                        ):
                            message.oscillation_angle = 3  # 90
                        else:
                            message.oscillation_angle = (
                                MideaFADevice._oscillation_angles.index(
                                    self._attributes[DeviceAttributes.oscillation_angle]
                                )
                            )
                    elif value == "Tilting":
                        if self._attributes[DeviceAttributes.tilting_angle] == "Off":
                            message.tilting_angle = 3  # 90
                        else:
                            message.tilting_angle = MideaFADevice._tilting_angles.index(
                                self._attributes[DeviceAttributes.tilting_angle]
                            )
                    else:
                        if (
                            self._attributes[DeviceAttributes.oscillation_angle]
                            == "Off"
                        ):
                            message.oscillation_angle = 3  # 90
                        else:
                            message.oscillation_angle = (
                                MideaFADevice._oscillation_angles.index(
                                    self._attributes[DeviceAttributes.oscillation_angle]
                                )
                            )
                        if self._attributes[DeviceAttributes.tilting_angle] == "Off":
                            message.tilting_angle = 3  # 90
                        else:
                            message.tilting_angle = MideaFADevice._tilting_angles.index(
                                self._attributes[DeviceAttributes.tilting_angle]
                            )
            elif attr == DeviceAttributes.oscillation_angle and (
                value in MideaFADevice._oscillation_angles or not value
            ):
                message = MessageSet(self._protocol_version, self.subtype)
                if value == "Off" or not value:
                    if self._attributes[DeviceAttributes.tilting_angle] == "Off":
                        message.oscillate = False
                    else:
                        message.oscillate = True
                        message.oscillation_mode = 2
                        message.tilting_angle = MideaFADevice._tilting_angles.index(
                            self._attributes[DeviceAttributes.tilting_angle]
                        )
                else:
                    message.oscillation_angle = MideaFADevice._oscillation_angles.index(
                        value
                    )
                    message.oscillate = True
                    if self._attributes[DeviceAttributes.tilting_angle] == "Off":
                        message.oscillation_mode = 1
                    elif (
                        self._attributes[DeviceAttributes.oscillation_mode] == "Tilting"
                    ):
                        message.oscillation_mode = 6
                        message.tilting_angle = MideaFADevice._tilting_angles.index(
                            self._attributes[DeviceAttributes.tilting_angle]
                        )
            elif attr == DeviceAttributes.tilting_angle and (
                value in MideaFADevice._tilting_angles or not value
            ):
                message = MessageSet(self._protocol_version, self.subtype)
                if value == "Off" or not value:
                    if self._attributes[DeviceAttributes.oscillation_angle] == "Off":
                        message.oscillate = False
                    else:
                        message.oscillate = True
                        message.oscillation_mode = 1
                        message.oscillation_angle = (
                            MideaFADevice._oscillation_angles.index(
                                self._attributes[DeviceAttributes.oscillation_angle]
                            )
                        )
                else:
                    message.tilting_angle = MideaFADevice._tilting_angles.index(value)
                    message.oscillate = True
                    if self._attributes[DeviceAttributes.oscillation_angle] == "Off":
                        message.oscillation_mode = 2
                    elif (
                        self._attributes[DeviceAttributes.oscillation_mode]
                        == "Oscillation"
                    ):
                        message.oscillation_mode = 6
                        message.oscillation_angle = (
                            MideaFADevice._oscillation_angles.index(
                                self._attributes[DeviceAttributes.oscillation_angle]
                            )
                        )
        return message

    def set_attribute(self, attr, value):
        message = None
        if attr in [
            DeviceAttributes.oscillate,
            DeviceAttributes.oscillation_mode,
            DeviceAttributes.oscillation_angle,
            DeviceAttributes.tilting_angle,
        ]:
            message = self.set_oscillation(attr, value)
        elif (
            attr == DeviceAttributes.fan_speed
            and value > 0
            and not self._attributes[DeviceAttributes.power]
        ):
            message = MessageSet(self._protocol_version, self.subtype)
            message.fan_speed = value
            message.power = True
        elif attr == DeviceAttributes.mode:
            if value in MideaFADevice._modes:
                message = MessageSet(self._protocol_version, self.subtype)
                message.mode = MideaFADevice._modes.index(value)
        elif not (attr == DeviceAttributes.fan_speed and value == 0):
            message = MessageSet(self._protocol_version, self.subtype)
            setattr(message, str(attr), value)
        if message is not None:
            self.build_send(message)

    def turn_on(self, fan_speed=None, mode=None):
        message = MessageSet(self._protocol_version, self.subtype)
        message.power = True
        if fan_speed is not None:
            message.fan_speed = fan_speed
        if mode is None:
            message.mode = mode
        self.build_send(message)

    def set_customize(self, customize):
        self._speed_count = self._default_speed_count
        if customize and len(customize) > 0:
            try:
                params = json.loads(customize)
                if params and "speed_count" in params:
                    self._speed_count = params.get("speed_count")
            except Exception as e:
                _LOGGER.error(f"[{self.device_id}] Set customize error: {repr(e)}")
            self.update_all({"speed_count": self._speed_count})


class MideaAppliance(MideaFADevice):
    pass



from ...message import MessageType, MessageRequest, MessageResponse, MessageBody


class MessageB4Base(MessageRequest):
    def __init__(self, protocol_version, message_type, body_type):
        super().__init__(
            device_type=0xB4,
            protocol_version=protocol_version,
            message_type=message_type,
            body_type=body_type,
        )

    @property
    def _body(self):
        raise NotImplementedError


class MessageQuery(MessageB4Base):
    def __init__(self, protocol_version):
        super().__init__(
            protocol_version=protocol_version,
            message_type=MessageType.query,
            body_type=0x01,
        )

    @property
    def _body(self):
        return bytearray([])


class B4MessageBody(MessageBody):
    def __init__(self, body):
        super().__init__(body)
        self.time_remaining = (
            (0 if body[22] == 0xFF else body[22]) * 3600
            + (0 if body[23] == 0xFF else body[23]) * 60
            + (0 if body[24] == 0xFF else body[24])
        )
        self.current_temperature = (body[25] << 8) + body[26]
        if self.current_temperature == 0:
            self.current_temperature = (body[27] << 8) + body[28]
        self.status = body[31]
        self.door = (body[32] & 0x02) > 0
        self.tank_ejected = (body[16] & 0x04) > 0
        self.water_shortage = (body[16] & 0x08) > 0
        self.water_change_reminder = (body[16] & 0x10) > 0


class MessageB4Response(MessageResponse):
    def __init__(self, message):
        super().__init__(message)
        if self.message_type in [
            MessageType.notify1,
            MessageType.query,
            MessageType.set,
        ]:
            if self.body_type == 0x01:
                self.set_body(B4MessageBody(super().body))
        self.set_attr()
